let eq_lcc = "cstr.eq"

open Util

open Kernel

open Big_int

module T = struct
  type 'a t =
    | Const of Pos.range * Const.t
    | Kernel_const of Pos.range * Const.t
    | Compound of Pos.range * 'a Compound.t
    | Record of Pos.range * bool * 'a String.Map.t
    | Nil of Pos.range
    | Cons of Pos.range * 'a * 'a
    | List of Pos.range * 'a list
    | Concat of Pos.range * 'a * 'a

  let map f t =
    match t with
      | Const (p, c) -> Const (p, c)
      | Kernel_const (p, c) -> Kernel_const (p, c)
      | Compound (p, c) -> Compound (p, Compound.map f c)
      | Record (p, closed, m) -> Record (p, closed, String.Map.map f m)
      | Nil p -> Nil p
      | Cons (p, v0, v1) -> Cons (p, f v0, f v1)
      | List (p, l) -> List (p, List.map f l)
      | Concat (p, v0, v1) -> Concat (p, f v0, f v1)

  let fold f v t =
    match t with
      | Const _
      | Kernel_const _
      | Nil _ -> v
      | Compound (_, c) -> Compound.fold_left f v c
      | Record (_, _, m) -> String.Map.fold (fun _ value v -> f v value) m v
      | List (_, l) -> List.fold_left f v l
      | Cons (_, v0, v1) -> f (f v v0) v1
      | Concat (_, v0, v1) -> f (f v v0) v1
end

module V = struct
  type t =
    | Underscore
    | Var of Var.t
    | Term of t T.t

  let rec fv v =
    match v with
    | Underscore -> Var.Set.empty
    | Var v -> Var.Set.singleton v
    | Term t -> T.fold (fun s v -> Var.Set.union (fv v) s) Var.Set.empty t
end

module Expr = struct
  type 'a t =
    | Underscore of Pos.range
    | Var of Pos.range * Var.t
    | Const of Pos.range * Const.t
    | Kernel_const of Pos.range * Const.t
    | Fun of Pos.range * 'a t * 'a t Compound.t
    | Compound of Pos.range * 'a t Compound.t
    | Record of Pos.range * bool * 'a t String.Map.t
    | Exists of Pos.range * Var.t * 'a t
    | Hide of Pos.range * Var.t * 'a t
    | Not of Pos.range * 'a t
    | Or of Pos.range * 'a t * 'a t
    | And of Pos.range * 'a t * 'a t
    | Eq of Pos.range * 'a t * 'a t
    | Neq of Pos.range * 'a t * 'a t
    | If of Pos.range * Var.Set.t * 'a t * 'a t * 'a t
    | Module of Pos.range * Var.t * 'a
    | Let of Pos.range * Var.t * 'a t * 'a t
    | Where of Pos.range * 'a t * 'a
    | Nil of Pos.range
    | Cons of Pos.range * 'a t * 'a t
    | List of Pos.range * 'a t list
    | Concat of Pos.range * 'a t * 'a t

  let rec of_val pos v =
    match v with
      | V.Underscore -> Underscore pos
      | V.Var v -> Var (pos, v)
      | V.Term (T.Const (pos, c)) -> Const (pos, c)
      | V.Term (T.Kernel_const (pos, c)) -> Kernel_const (pos, c)
      | V.Term (T.Compound (pos, c)) ->
          Compound (pos, Compound.map (of_val pos) c)
      | V.Term (T.Record (pos, closed, m)) ->
          Record (pos, closed, String.Map.map (of_val pos) m)
      | V.Term (T.Nil pos) ->
          Nil pos
      | V.Term (T.Cons (pos, e0, e1)) ->
          Cons (pos, of_val pos e0, of_val pos e1)
      | V.Term (T.List (pos, l)) ->
          List (pos, List.map (of_val pos) l)
      | V.Term (T.Concat (pos, e0, e1)) ->
          Concat (pos, of_val pos e0, of_val pos e1)

  let of_big_int pos i = Const (pos, Const.Int i)

  let of_int pos i = Const (pos, Const.Int (big_int_of_int i))

  let rec fv fv_a e =
    match e with
      | Underscore _
      | Const _
      | Kernel_const _
      | Nil _ -> Var.Set.empty
      | Var (_, v) -> Var.Set.singleton v
      | Fun (_, e, c) -> Var.Set.union (fv fv_a e) (fv_compound fv_a c)
      | Compound (_, c) -> fv_compound fv_a c
      | Record (_, _, m) ->
          let add _ value s = Var.Set.union (fv fv_a value) s in
          String.Map.fold add m Var.Set.empty
      | Exists (_, x, e) -> Var.Set.remove x (fv fv_a e)
      | Hide (_, _, e)
      | Not (_, e) -> fv fv_a e
      | Or (_, e0, e1)
      | And (_, e0, e1)
      | Eq (_, e0, e1)
      | Neq (_, e0, e1) -> Var.Set.union (fv fv_a e0) (fv fv_a e1)
      | If (_, vars, cond, case_true, case_false) ->
          let v = Var.Set.union (fv fv_a cond) (fv fv_a case_true) in
          let v = Var.Set.diff v vars in
          Var.Set.union v (fv fv_a case_false)
      | Module (_, x, e) -> Var.Set.remove x (fv_a e)
      | Let (_, x, v, e) ->
          Var.Set.union (fv fv_a v) (Var.Set.remove x (fv fv_a e))
      | Where (_, e, a) -> Var.Set.union (fv fv_a e) (fv_a a)
      | Cons (_, e0, e1) -> Var.Set.union (fv fv_a e0) (fv fv_a e1)
      | List (_, l) ->
          let add s e = Var.Set.union s (fv fv_a e) in
          List.fold_left add Var.Set.empty l
      | Concat (_, e0, e1) -> Var.Set.union (fv fv_a e0) (fv fv_a e1)
  and fv_compound fv_a c =
    Compound.fold_left (fun s v -> Var.Set.union (fv fv_a v) s) Var.Set.empty c

  let lib_call path op pos args =
    let m_op = Var (pos, Files.use_module pos path) in
    let c = Compound.mk op (Array.of_list args) in
    Fun (pos, m_op, c)

  let rec big_exists p l a =
    match l with
      | [] -> a
      | h :: t -> Exists (p, h, big_exists p t a)
end

type t =
  | Skip
  | Tok of Pos.range * bool * expr * expr Compound.t
  | Eq of Pos.range * expr * expr
  | Call of Pos.range * expr Compound.t
  | Ask of Pos.range * expr * Var.t Compound.t * t
  | PersistentAsk of Pos.range * Var.Set.t * (Pos.range * expr) list * t
  | TransientAsk of Pos.range * Var.Set.t * (Pos.range * expr) list * t
  | If of Pos.range * Var.Set.t * expr * t * t
  | Exists of Pos.range * Var.t * t
  | Hide of Pos.range * Var.t * t
  | Par of t * t
  | Let of Pos.range * Var.t * expr * t
and expr = t Expr.t

let lib_call path op pos args =
  let m_op = Expr.Var (pos, Files.use_module pos path) in
  let c = Compound.mk op (Array.of_list args) in
  Tok (pos, true, m_op, c)

let rec fv a =
  match a with
    | Skip -> Var.Set.empty
    | Tok (_, _, e, c) -> Var.Set.union (Expr.fv fv e) (Expr.fv_compound fv c)
    | Eq (_, e0, e1) -> Var.Set.union (Expr.fv fv e0) (Expr.fv fv e1)
    | Call (_, c) -> Expr.fv_compound fv c
    | Ask (_, e, c, a) ->
        Var.Set.union (Expr.fv fv e)
          (Compound.fold_left (fun s x -> Var.Set.remove x s) (fv a) c)
    | PersistentAsk (_, v, g, a)
    | TransientAsk (_, v, g, a) ->
        let vg =
          List.fold_left (fun v (_, g) -> Var.Set.union (Expr.fv fv g) v)
            Var.Set.empty g in
        Var.Set.diff (Var.Set.union vg (fv a)) v
    | If (_, vars, cond, case_true, case_false) ->
        let v = Var.Set.union (Expr.fv fv cond) (fv case_true) in
        let v = Var.Set.diff v vars in
        Var.Set.union v (fv case_false)
    | Exists (_, x, a) -> Var.Set.remove x (fv a)
    | Hide (_, _, a) -> fv a
    | Par (a0, a1) -> Var.Set.union (fv a0) (fv a1)
    | Let (_, x, e, a) -> Var.Set.union (Expr.fv fv e) (Var.Set.remove x (fv a))

let rec big_par l =
  match l with
    | [] -> Skip
    | [a] -> a
    | h :: t -> Par (h, big_par t)

let rec big_exists p l a =
  match l with
    | [] -> a
    | h :: t -> Exists (p, h, big_exists p t a)

let rec big_exists_pos l a =
  match l with
    | [] -> a
    | (p, h) :: t -> Exists (p, h, big_exists_pos t a)

let rec big_let p l a =
  match l with
    | [] -> a
    | (x, v) :: t -> Let (p, x, v, big_let p t a)

type point = (fresh, value) Unionfind.point
and fresh = {
  p: Pos.range;
  mutable v: Var.t option;
}
and value =
  | Term of point T.t
  | ExternVar of Var.t

let fresh p = Unionfind.fresh { p = p; v = None }

let of_term t = Unionfind.of_value (Term t)

let of_extern_var v = Unionfind.of_value (ExternVar v)

type expr_env = {
  parent: Var.t -> point;
  mutable vars: (Pos.range * Var.t) list;
  mutable eqs: (Pos.range * point * point) list;
  mutable toks: (Pos.range * bool * point * point Compound.t) list;
}

let fresh_to_var env pos f =
  match f.v with
    | None ->
        let v = Var.fresh () in
        env.vars <- (pos, v) :: env.vars;
        f.v <- Some v;
        v
    | Some v -> v

let rec of_extern_point env pos pt =
  let (_, d) = Unionfind.find pt in
  match d with
    | Unionfind.Fresh f -> of_extern_var (fresh_to_var env pos f)
    | Unionfind.Value (Term t) -> of_term (T.map (of_extern_point env pos) t)
    | Unionfind.Value (ExternVar v) -> of_extern_var v

let create_expr_env parent = { parent = parent; vars = []; eqs = []; toks = [] }

let unify env pos v0 v1 =
  let rec unify_term t0 t1 =
    match t0, t1 with
      | T.Const (_, c0), T.Const (_, c1) when Const.equal c0 c1 -> ()
      | T.Compound (_, c0), T.Compound (_, c1)
          when Sgn.equal c0.Compound.s c1.Compound.s ->
          let args0 = c0.Compound.args and args1 = c1.Compound.args in
          for i = 0 to Array.length args0 - 1 do
            Unionfind.union unify_value pos args0.(i) args1.(i)
          done;
          ()
      | _ -> raise Unionfind.Fail
  and unify_value v0 v1 =
    match v0, v1 with
      | Term t0, Term t1 -> unify_term t0 t1
      | _ ->
          let v0 = Unionfind.of_value v0 in
          let v1 = Unionfind.of_value v1 in
          env.eqs <- (pos, v0, v1) :: env.eqs in
  try Unionfind.union unify_value pos v0 v1
  with Unionfind.Fail -> Err.constraint_always_inconsistent pos

let rec get_val env ctx p v =
  match v with
    | V.Underscore -> fresh p
    | V.Var v ->
        begin
          try String.Map.find v ctx
          with Not_found -> env.parent v
        end
    | V.Term t -> of_term (T.map (get_val env ctx p) t)

let intro p ctx x =
  let x' = fresh p in
  String.Map.add x x' ctx

let call_fun env p m f args =
  let r = fresh p in
  let t = Compound.mk f (Array.of_list (args @ [r])) in
  env.toks <- (p, true, m, t) :: env.toks;
  r

let bool_and env pos e0 e1 =
  let m = of_extern_var (Files.use_module pos ["kernel"; "bool"]) in
  call_fun env pos m "and" [e0; e1]

let bool_or env pos e0 e1 =
  let m = of_extern_var (Files.use_module pos ["kernel"; "bool"]) in
  call_fun env pos m "or" [e0; e1]

let bool_not env pos e =
  let m = of_extern_var (Files.use_module pos ["kernel"; "bool"]) in
  call_fun env pos m "not" [e]

let arith_eq env pos e0 e1 =
  let m = of_extern_var (Files.use_module pos ["data"; "term"]) in
  call_fun env pos m "eq" [e0; e1]

let arith_neq env pos e0 e1 =
  let m = of_extern_var (Files.use_module pos ["data"; "term"]) in
  call_fun env pos m "ne" [e0; e1]

let new_parent env ctx pos v =
  try of_extern_point env pos (String.Map.find v ctx)
  with Not_found -> env.parent v

let check_guarded pos vars guarded =
  let diff = Var.Set.diff vars guarded in
  if not (Var.Set.is_empty diff) then
    Err.unguarded_variables pos diff;

module If = struct
  let value_var var = "value_" ^ var

  let compound_var x e = Compound.mk (value_var x) [| e |]

  let ask_vars consume pos ex vars body =
    let ask_var var body =
      let transient = Var.fresh () in
      let body =
        if consume then body
        else 
          Par (
            Tok (
              pos, false, ex, compound_var var (Expr.Var (pos, var))), body) in
      Exists (
        pos, transient,
        Par (
          Tok (
            pos, false, Expr.Var (pos, transient),
            Compound.mk "transient" [| |]),
          Ask (
            pos, ex, compound_var var var,
            Ask (
              pos, Expr.Var (pos, transient), Compound.mk "transient" [| |],
              body)))) in
    Var.Set.fold ask_var vars body

  let return pos vars ev ec e a =
    let body = Par (Eq (pos, ec, e), a) in
    let unsafe = Var.Set.inter vars (Expr.fv fv e) in
    ask_vars false pos ev unsafe body

  let simple_if pos cond case_true case_false =
    let x = Var.fresh () in
    let ex = Expr.Var (pos, x) in
    let ask_true = Ask (pos, ex, Compound.mk "true" [| |], case_true) in
    let ask_false = Ask (pos, ex, Compound.mk "false" [| |], case_false) in
    Let (pos, x, cond, Par (ask_true, ask_false))

  let big_op one neutral pos result conds =
    let tell_one = Tok (pos, false, result, Compound.mk one [| |]) in
    let tell_neutral = Tok (pos, false, result, Compound.mk neutral [| |]) in
    let ask_one body cond = Ask (pos, cond, Compound.mk one [| |], body) in
    let p = Var.fresh () in
    let ep = Expr.Var (pos, p) in
    let ask_neutral cond =
      Ask (
        pos, cond, Compound.mk neutral [| |],
        Ask (
          pos, Expr.Fun (pos, ep, Compound.mk "check" [| |]),
          Compound.mk "done" [| |], tell_neutral)) in
    let one_agent = List.fold_left ask_one tell_one conds in
    let neutral_agent = big_par (List.map ask_neutral conds) in
    Let (
      pos, p, Expr.lib_call ["control"; "point"] "new" pos [],
      Par (one_agent, neutral_agent))

  let conj = big_op "true" "false"

  let disj = big_op "false" "true"

  let check_type pos e ty ec body =
    let etype = Expr.Fun (pos, e, Compound.mk "type" [| |]) in
    let id = Expr.of_int pos (Sgn.get ty 0).Sgn.id in
    let cond_type = Expr.lib_call ["kernel"; "int"] "eq" pos [etype; id] in
    let case_false = Tok (pos, false, ec, Compound.mk "false" [| |]) in
    simple_if pos cond_type body case_false

  let eq_array make_eq pos vars ev ec unguarded a0 a1 =
    let arity = Array.length a0 in
    let rec mk_arg index unguarded =
      if index < arity then
        match a0.(index), a1.(index) with
          | Expr.Underscore _, Expr. Underscore _ ->
              mk_arg (succ index) unguarded
          | arg0, arg1 ->
              let c = Var.fresh () in
              let ec = Expr.Var (pos, c) in
              let s, a = make_eq pos vars ev ec unguarded arg0 arg1 in
              let s', c', a' =
                mk_arg (succ index) (Var.Set.diff unguarded s) in
              Var.Set.union s s', c :: c', Par (a, a')
      else Var.Set.empty, [], Skip in
    let s, conds, case_true = mk_arg 0 unguarded in
    s, big_exists pos conds (
      Par (
        case_true,
        conj pos ec (List.map (fun v -> Expr.Var (pos, v)) conds)))

  let eq_compound make_eq pos vars ev ec unguarded c0 c1 =
    if not (Sgn.equal c0.Compound.s c1.Compound.s) then
      Err.constraint_always_inconsistent pos;
    eq_array make_eq pos vars ev ec unguarded c0.Compound.args c1.Compound.args

  let eq_compound_expr make_eq pos vars ev ec unguarded c e =
    let x = Var.fresh () in
    let ex = Expr.Var (pos, x) in
    let csid = Expr.of_int pos (Compound.sid c) in
    let esid = Expr.Fun (pos, ex, Compound.mk "sig" [| |]) in
    let cond_sgn =
      Expr.lib_call ["kernel"; "int"] "eq" pos [csid; esid] in
    let arity = Compound.arity c in
    let rec mk_arg index unguarded =
      if index < arity then
        match Compound.get c index with
          | Expr.Underscore _ -> mk_arg (succ index) unguarded
          | arg ->
              let i = Expr.of_int pos index in
              let arg' = Expr.Fun (pos, ex, Compound.mk "get" [| i |]) in
              let c = Var.fresh () in
              let ec = Expr.Var (pos, c) in
              let s, a = make_eq pos vars ev ec unguarded arg arg' in
              let s', c', a' =
                mk_arg (succ index) (Var.Set.diff unguarded s) in
              Var.Set.union s s', c :: c', Par (a, a')
      else Var.Set.empty, [], Skip in
    let s, conds, case_true = mk_arg 0 unguarded in
    let case_true =
      big_exists pos conds (
        Par (
          case_true,
          conj pos ec (List.map (fun v -> Expr.Var (pos, v)) conds))) in
    let case_false = Tok (pos, false, ec, Compound.mk "false" [| |]) in
    let body = simple_if pos cond_sgn case_true case_false in
    s, Let (pos, x, e, check_type pos ex "compound" ec body)

  let eq_cons_expr make_eq pos vars ev ec unguarded hd tl e =
    let x = Var.fresh () in
    let ex = Expr.Var (pos, x) in
    let hd' = Expr.Fun (pos, ex, Compound.mk "hd" [| |]) in
    let hdc = Var.fresh () in
    let ehdc = Expr.Var (pos, hdc) in
    let shd, ahd = make_eq pos vars ev ehdc unguarded hd hd' in
    let tl' = Expr.Fun (pos, ex, Compound.mk "tl" [| |]) in
    let tlc = Var.fresh () in
    let etlc = Expr.Var (pos, tlc) in
    let stl, atl = make_eq pos vars ev etlc unguarded tl tl' in
    let s = Var.Set.union shd stl in
    let body =
      big_exists pos [hdc; tlc]
        (big_par [ahd; atl; conj pos ec [ehdc; etlc]]) in
    s, Let (pos, x, e, check_type pos ex "list" ec body)

  let eq_record make_eq pos vars ev ec unguarded closed0 m0 closed1 m1 =
    let check_presence field _ =
      if not (String.Map.mem field m0) then
        Err.constraint_always_inconsistent pos in
    if closed0 then
      String.Map.iter check_presence m1;
    let rec mk_field field v0 (s, c, a, unguarded) =
      try
        let v1 = String.Map.find field m1 in
        match v0, v1 with
          | Expr.Underscore _, Expr. Underscore _ ->
              (s, c, a, unguarded)
          | _ ->
              let c' = Var.fresh () in
              let ec = Expr.Var (pos, c') in
              let s', a' = make_eq pos vars ev ec unguarded v0 v1 in
              Var.Set.union s s', c' :: c, Par (a, a'),
              Var.Set.diff unguarded s
      with Not_found ->
        if closed1 then
          Err.constraint_always_inconsistent pos
        else
          (s, c, a, unguarded) in
    let s, conds, case_true, _ =
      String.Map.fold mk_field m0 (Var.Set.empty, [], Skip, unguarded) in
    s, big_exists pos conds (
      Par (
        case_true,
        conj pos ec (List.map (fun v -> Expr.Var (pos, v)) conds)))

  let eq_record_expr make_eq pos vars ev ec unguarded closed m e =
    let x = Var.fresh () in
    let ex = Expr.Var (pos, x) in
    let rec mk_field field value (s, c, a, unguarded) =
      let i = Expr.of_int pos (Sgn.get field 0).Sgn.id in
      let has_field = Expr.Fun (pos, ex, Compound.mk "has_field" [| i |]) in
      let c' = Var.fresh () in
      let ec = Expr.Var (pos, c') in
      let case_false = Tok (pos, false, ec, Compound.mk "false" [| |]) in
      let s, case_true, unguarded =
        match value with
          | Expr.Underscore _ ->
              s, Tok (pos, false, ec, Compound.mk "true" [| |]), unguarded
          | _ ->
              let value' = Expr.Fun (pos, ex, Compound.mk field [| |]) in
              let s', a' = make_eq pos vars ev ec unguarded value value' in
              Var.Set.union s s', a', Var.Set.diff unguarded s' in
      let a' = simple_if pos has_field case_true case_false in
      s, c' :: c, Par (a, a'), unguarded in
    let s, conds, case_true, _ =
      String.Map.fold mk_field m (Var.Set.empty, [], Skip, unguarded) in
    let case_true =
      big_exists pos conds (
        Par (
          case_true,
          conj pos ec (List.map (fun v -> Expr.Var (pos, v)) conds))) in
    let body =
      if closed then
        let count _ _ i = succ i in
        let mfcount = Expr.of_int pos (String.Map.fold count m 0) in
        let efcount = Expr.Fun (pos, ex, Compound.mk "field_count" [| |]) in
        let cond_sgn =
          Expr.lib_call ["kernel"; "int"] "eq" pos [mfcount; efcount] in
        let case_false = Tok (pos, false, ec, Compound.mk "false" [| |]) in
        simple_if pos cond_sgn case_true case_false
      else
        case_true in
    s, Let (pos, x, e, check_type pos ex "record" ec body)

  exception Not_a_list

  let rec to_list e =
    match e with
    | Expr.List (_, l) -> l
    | Expr.Concat (_, e0, e1) -> to_list e0 @ to_list e1
    | _ -> raise Not_a_list

  let cons pos hd tl = Expr.Cons (pos, hd, tl)

  let rec canonize e =
    match e with
      | Expr.List (pos, l) ->
          List.fold_right (cons pos) l (Expr.Nil pos)
      | Expr.Concat (pos, e0, e1) ->
          let e1 = canonize e1 in
          begin
            try
              let l = to_list e0 in
              List.fold_right (cons pos) l e1
            with Not_a_list ->
              let e0 = canonize e0 in
              Expr.Concat (pos, e0, e1)
          end
      | _ -> e

  let rec make_eq pos vars ev ec unguarded e0 e1 =
    match canonize e0, canonize e1 with
      | Expr.Underscore pos, e
      | e, Expr.Underscore pos ->
          Var.Set.empty,
          Par (
            return pos vars ev (Expr.Underscore pos) e Skip,
            Tok (pos, false, ec, Compound.mk "true" [| |]))
      | Expr.Var (pos, x), e
      | e, Expr.Var (pos, x) when
          Var.Set.mem x unguarded && not (Var.Set.mem x (Expr.fv fv e)) ->
            Var.Set.singleton x,
            Par (
              Tok (pos, false, ev, compound_var x e),
              Tok (pos, false, ec, Compound.mk "true" [| |]))
      | Expr.Nil _, Expr.Cons _
      | Expr.Cons _, Expr.Nil _ ->
          Err.constraint_always_inconsistent pos;
      | Expr.Nil _, Expr.Nil _ ->
          Var.Set.empty,
          Tok (pos, false, ec, Compound.mk "true" [| |])
      | Expr.Cons (_, hd0, tl0), Expr.Cons (_, hd1, tl1) ->
          eq_array make_eq pos vars ev ec unguarded
            [| hd0; tl0 |] [| hd1; tl1 |]
      | Expr.Cons (_, hd, tl), e
      | e, Expr.Cons (_, hd, tl) ->
          let x = Var.fresh () in
          let ex = Expr.Var (pos, x) in
          let s, body =
            eq_cons_expr make_eq pos vars ev ec unguarded hd tl ex in
          s, Exists (pos, x, return pos vars ev ex e body)
      | Expr.Compound (_, c0), Expr.Compound (_, c1) ->
          eq_compound make_eq pos vars ev ec unguarded c0 c1
      | Expr.Compound (pos, c), e | e, Expr.Compound (pos, c) ->
          let x = Var.fresh () in
          let ex = Expr.Var (pos, x) in
          let s, body = 
            if Compound.arity c = 0 then
              let csid = Expr.of_int pos (Compound.sid c) in
              let body = Expr.lib_call ["kernel"; "int"] "eq" pos [csid; ex] in
              let body = check_type pos ex "int" ec (Eq (pos, ec, body)) in
              Var.Set.empty, body
            else
              eq_compound_expr make_eq pos vars ev ec unguarded c ex in
          s, Exists (pos, x, return pos vars ev ex e body)
      | Expr.Record (_, closed0, m0), Expr.Record (_, closed1, m1) ->
          eq_record make_eq pos vars ev ec unguarded closed0 m0 closed1 m1
      | Expr.Record (pos, closed, m), e | e, Expr.Record (pos, closed, m) ->
          let x = Var.fresh () in
          let ex = Expr.Var (pos, x) in
          let s, body =
            eq_record_expr make_eq pos vars ev ec unguarded closed m ex in
          s, Exists (pos, x, return pos vars ev ex e body)
      | _ ->
          let body = Expr.lib_call ["data"; "term"] "eq" pos [e0; e1] in
          Var.Set.empty, return pos vars ev ec body Skip

  let rec make pos vars ev ec unguarded cond =
    match cond with
      | Expr.Eq (pos, e0, e1) ->
          make_eq pos vars ev ec unguarded e0 e1
      | Expr.And (pos, e0, e1) ->
          let c0 = Var.fresh () in
          let ec0 = Expr.Var (pos, c0) in
          let s, a0 = make pos vars ev ec0 unguarded e0 in
          let c1 = Var.fresh () in
          let ec1 = Expr.Var (pos, c1) in
          let s', a1 = make pos vars ev ec1 (Var.Set.diff unguarded s) e1 in
          Var.Set.union s s',
          Exists (
            pos, c0,
            Exists (pos, c1, Par (Par (a0, a1), conj pos ec [ec0; ec1])))
      | Expr.Or (pos, e0, e1) ->
          let v0 = Var.fresh () in
          let ev0 = Expr.Var (pos, v0) in
          let c0 = Var.fresh () in
          let ec0 = Expr.Var (pos, c0) in
          let s, a0 = make pos vars ev0 ec0 unguarded e0 in
          let v1 = Var.fresh () in
          let ev1 = Expr.Var (pos, v1) in
          let c1 = Var.fresh () in
          let ec1 = Expr.Var (pos, c1) in
          let s', a1 = make pos vars ev1 ec1 unguarded e1 in
          if not (Var.Set.equal s s') then
            Err.some_variables_only_appear_in_one_branch
              pos (Var.Set.union (Var.Set.diff s s') (Var.Set.diff s' s));
          let tell_one = Tok (pos, false, ec, Compound.mk "false" [| |]) in
          let tell_neutral = Tok (pos, false, ec, Compound.mk "true" [| |]) in
          let ask_one body cond =
            Ask (pos, cond, Compound.mk "false" [| |], body) in
          let ask_neutral ev' cond =
            let add_tell_v var body =
              Par 
                (body,
                 Tok (
                   pos, false, ev,
                   compound_var var (Expr.Var (pos, var)))) in
            let body = Var.Set.fold add_tell_v s tell_neutral in
            let add_ask_v var body =
              Ask (pos, ev', compound_var var var, body) in
            let body = Var.Set.fold add_ask_v s body in
            Ask (pos, cond, Compound.mk "true" [| |], body) in
          let one_agent = ask_one (ask_one tell_one ec0) ec1 in
          let neutral_agent = Par (ask_neutral ev0 ec0, ask_neutral ev1 ec1) in
          let disj = Par (one_agent, neutral_agent) in
          let body = Par (Par (a0, a1), disj) in
          s,
          Exists (
            pos, v0, Exists (pos, c0, Exists (pos, v1, Exists (pos, c1, body))))
      | _ ->
          Var.Set.empty, return pos vars ev ec cond Skip

  let compile pos vars cond case_true case_false =
    let x = Var.fresh () in
    let ex = Expr.Var (pos, x) in
    let guarded, check = make pos vars ex ex vars cond in
    check_guarded pos vars guarded;
    let case_true = ask_vars true pos ex vars case_true in
    Exists (pos, x, Par (check, simple_if pos ex case_true case_false))
end

module Ask = struct
  let compile_tok pos vars e c =
    let unguarded = Var.Set.inter (Expr.fv fv e) vars in
    if not (Var.Set.is_empty unguarded) then
      Err.unguarded_variables pos unguarded;
    let add_arg (arg_list, eq_list, vars) arg =
      match arg with
        | Expr.Var (_, v) when Var.Set.mem v vars ->
            (v :: arg_list, eq_list, Var.Set.remove v vars)
        | _ ->
            let v = Var.fresh () in
            (v :: arg_list, (Expr.Var (pos, v), arg) :: eq_list, vars) in
    let arg_list, eq_list, vars = Compound.fold_left add_arg ([], [], vars) c in
    let arg_list = List.rev arg_list in
    let f = Compound.f c in
    let args = Array.of_list arg_list in
    match eq_list with
      | [] -> (fun body -> Ask (pos, e, Compound.mk f args, body)), vars
      | h :: t ->
          let mk_eq (e0, e1) = Expr.Eq (pos, e0, e1) in
          let add_cond cond eq = Expr.And (pos, mk_eq eq, cond) in
          let cond = List.fold_left add_cond (mk_eq h) t in
          let x = Var.fresh () in
          let ex = Expr.Var (pos, x) in
          let guarded, check = If.make pos vars ex ex vars cond in
          let vars = Var.Set.diff vars guarded in
          (fun body ->
             let body = If.ask_vars true pos ex guarded body in
             let body =
               Exists (pos, x, Par (check, If.simple_if pos ex body Skip)) in
             Ask (pos, e, Compound.mk f args, body)),
          vars

  let compile_guard pos vars guard =
    match guard with
      | Expr.Fun (pos, e, c) ->
          compile_tok pos vars e c
      | _ ->
          let x = Var.fresh () in
          let ex = Expr.Var (pos, x) in
          let guarded, check = If.make pos vars ex ex vars guard in
          let vars = Var.Set.diff vars guarded in
          (fun body ->
             let body = If.ask_vars true pos ex guarded body in
             Exists (pos, x, Par (check, If.simple_if pos ex body Skip))),
          vars

  let rec compile_list pos vars guard body =
    match guard with
      | [] ->
          if not (Var.Set.is_empty vars) then
            Err.unguarded_variables pos vars;
          body
      | (pos, h) :: t ->
          let context, vars = compile_guard pos vars h in
          context (compile_list pos vars t body)

  let compile_transient pos vars guard body =
    match guard with
      | [] ->
          if not (Var.Set.is_empty vars) then
            Err.unguarded_variables pos vars;
          body
      | _ ->
          let transient = Var.fresh () in
          let etransient = Expr.Var (pos, transient) in
          let body = Ask (pos, etransient, Compound.mk "transient" [| |], body) in
          let result = compile_list pos vars guard body in
          let tok_transient =
            Tok (pos, false, etransient, Compound.mk "transient" [| |]) in
          Exists (pos, transient, Par (tok_transient, result))

  let compile_persistent pos vars guard body =
    match guard with
      | [] -> Err.persistent_ask_guard_cannot_be_empty pos
      | (pos, h) :: t ->
          begin
            match h with
              | Expr.Fun (pos, e, c) ->
                  let context, vars = compile_tok pos vars e c in
                  context (compile_transient pos vars t body)
              | _ ->
                  Err.persistent_ask_guard_should_begin_by_a_token pos
          end
end

type env = {
  expr: expr_env;
  mutable calls: (Pos.range * point Compound.t) list;
  mutable asks:
    (point String.Map.t * Pos.range * point * Var.t Compound.t * t) list;
  mutable internal_agents: Eq.t list;
}

let rec eval env ctx e =
  match e with
    | Expr.Underscore pos -> fresh pos
    | Expr.Var (pos, v) ->
        begin
          try String.Map.find v ctx
          with Not_found -> env.expr.parent v
        end
    | Expr.Const (pos, c) -> of_term (T.Const (pos, c))
    | Expr.Kernel_const (pos, c) -> of_term (T.Kernel_const (pos, c))
    | Expr.Fun (pos, m, t) ->
        let m = eval env ctx m in
        let args = List.map (eval env ctx) (Array.to_list t.Compound.args) in
        call_fun env.expr pos m (Compound.f t) args
    | Expr.Compound (pos, t) ->
        of_term (T.Compound (pos, Compound.map (eval env ctx) t))
    | Expr.Record (pos, closed, m) ->
        of_term (T.Record (pos, closed, String.Map.map (eval env ctx) m))
    | Expr.Exists (pos, x, e0) ->
        eval env (intro pos ctx x) e0
    | Expr.Hide (pos, x, e0) ->
        eval env (String.Map.remove x ctx) e0
    | Expr.Not (pos, e) ->
        bool_not env.expr pos (eval env ctx e)
    | Expr.And (pos, e0, e1) ->
        bool_and env.expr pos (eval env ctx e0) (eval env ctx e1)
    | Expr.Or (pos, e0, e1) ->
        bool_or env.expr pos (eval env ctx e0) (eval env ctx e1)
    | Expr.Eq (pos, e0, e1) ->
        arith_eq env.expr pos (eval env ctx e0) (eval env ctx e1)
    | Expr.Neq (pos, e0, e1) ->
        arith_neq env.expr pos (eval env ctx e0) (eval env ctx e1)
    | Expr.Module (pos, x, e) -> 
        let x' = fresh pos in
        split env (String.Map.add x x' ctx) e;
        x'
    | Expr.If (pos, vars, cond, case_true, case_false) ->
        let x = Var.fresh () in 
        let x' = fresh pos in
        let vx = Expr.Var (pos, x) in
        let a =
          If.compile pos vars cond
            (Eq (pos, vx, case_true)) (Eq (pos, vx, case_false)) in
        split env (String.Map.add x x' ctx) a;
        x'
    | Expr.Let (pos, x, e0, e1) ->
        let x' = fresh pos in
        let v0 = eval env ctx e0 in
        let v1 = eval env (String.Map.add x x' ctx) e1 in
        unify env.expr pos x' v0;
        v1
    | Expr.Where (pos, e, a) ->
        split env ctx a;
        eval env ctx e
    | Expr.Nil pos -> of_term (T.Nil pos)
    | Expr.Cons (pos, hd, tl) ->
        of_term (T.Cons (pos, eval env ctx hd, eval env ctx tl))
    | Expr.List (pos, l) -> of_term (T.List (pos, List.map (eval env ctx) l))
    | Expr.Concat (pos, e0, e1) ->
        let e0 = eval env ctx e0 in
        let e1 = eval env ctx e1 in
        of_term (T.Concat (pos, e0, e1))
and split env ctx a =
  match a with
    | Skip -> ()
    | Tok (pos, to_consume, m, t) ->
        let m' = eval env ctx m in
        let t' = Compound.map (eval env ctx) t in
        env.expr.toks <- (pos, to_consume, m', t') :: env.expr.toks
    | Eq (pos, v0, v1) ->
        unify env.expr pos (eval env ctx v0) (eval env ctx v1)
    | Call (pos, t) ->
        env.calls <- (pos, Compound.map (eval env ctx) t) :: env.calls
    | Ask (pos, m, t, a0) ->
        let m' = eval env ctx m in
        env.asks <- (ctx, pos, m', t, a0) :: env.asks
    | PersistentAsk (pos, vars, g, a) ->
        let a = Ask.compile_persistent pos vars g a in
        split env ctx a
    | TransientAsk (pos, vars, g, a) ->
        let a = Ask.compile_transient pos vars g a in
        split env ctx a
    | If (pos, vars, cond, case_true, case_false) ->
        let a = If.compile pos vars cond case_true case_false in
        split env ctx a
    | Exists (pos, x, a0) ->
        split env (intro pos ctx x) a0
    | Hide (pos, x, a0) ->
        split env (String.Map.remove x ctx) a0
    | Let (pos, x, e0, a1) ->
        let x' = fresh pos in
        let v0 = eval env ctx e0 in
        let a1 = split env (String.Map.add x x' ctx) a1 in
        unify env.expr pos x' v0;
        a1
    | Par (a0, a1) ->
        split env ctx a0;
        split env ctx a1

let call env p m f args =
  let v = Var.fresh () in
  env.expr.vars <- (p, v) :: env.expr.vars;
  let vv = Value.Var v in
  let m = Files.use_module p m in
  let args = Array.of_list (args @ [vv]) in
  let c = Compound.mk f args in
  env.internal_agents <-
    Eq.Tok (p, false, Value.Var m, c) :: env.internal_agents;
  vv

let compound env p t =
  if Compound.arity t >= 1 then
    let sid = Compound.sid t in
    call env p ["kernel"; "compound"] "make"
      [Value.of_int sid; Value.Term (Term.Array t.Compound.args)]
  else
    call env p ["kernel"; "int"] "make" [Value.of_int (Compound.sid t)]

let record sub env p closed l =
  let v = Var.fresh () in
  let vv = Value.Var v in
  env.expr.vars <- (p, v) :: env.expr.vars;
  let r = Var.fresh () in
  let vr = Value.Var r in
  let fields = ref [] in
  let values = ref [] in
  let make_field field value =
    let value = sub value in
    fields := Value.of_int (Sgn.get field 0).Sgn.id :: !fields;
    values := value :: !values;
    let body = Eq.Eq (p, vr, value) in
    let ask = Eq.Ask (p, vv, Compound.mk field [| r |], body) in
    env.internal_agents <- ask :: env.internal_agents in
  String.Map.iter make_field l;
  let fields_array = Array.of_list !fields in
  let fields_count = Value.of_int (Array.length fields_array) in
  let fields_array = Value.Term (Term.Array fields_array) in
  let values_array = Array.of_list !values in
  let values_array = Value.Term (Term.Array values_array) in
  let c =
    Compound.mk "make" [| vv; fields_count; fields_array; values_array |] in
  let m = Files.use_module p ["kernel"; "record"] in
  env.internal_agents <-
    Eq.Tok (p, false, Value.Var m, c) :: env.internal_agents;
  vv

let cons env pos hd tl =
  call env pos ["data"; "list"] "cons" [hd; tl]

let nil env pos =
  call env pos ["data"; "list"] "nil" []

let append sub env pos l t =
  List.fold_right (fun hd tl-> cons env pos (sub hd) tl) l t

let list sub env pos l =
  append sub env pos l (nil env pos)

exception Not_a_list

let rec to_list v =
  let (_, d) = Unionfind.find v in
  match d with
    | Unionfind.Value (Term t) ->
        begin
          match t with
            | T.List (_, l) -> l
            | T.Concat (_, v0, v1) ->
                to_list v0 @ to_list v1
            | _ -> raise Not_a_list
        end
    | _ -> raise Not_a_list

exception Not_a_string

let rec to_string t =
  match t with
    | T.Const (_, Const.String s) -> s
    | T.Concat (_, v0, v1) ->
        let (_, d0) = Unionfind.find v0 in
        let (_, d1) = Unionfind.find v1 in
        begin
          match d0, d1 with
            | Unionfind.Value (Term t0), Unionfind.Value (Term t1) ->
                to_string t0 ^ to_string t1
            | _ -> raise Not_a_string
        end
    | _ -> raise Not_a_string

let rec to_term sub env t =
  match t with
    | T.Const (pos, Const.Int i) ->
        call env pos ["kernel"; "int"] "make" [Value.of_big_int i]
    | T.Const (pos, Const.Float f) ->
        call env pos ["kernel"; "float"] "make"
          [Value.Term (Term.Const (Const.Float f))]
    | T.Kernel_const (pos, c) ->
        Value.Term (Term.Const c)
    | T.Compound (pos, t) ->
        if Compound.arity t >= 1 then
          compound env pos (Compound.map sub t)
        else
          call env pos ["kernel"; "int"] "make" [Value.of_int (Compound.sid t)]
    | T.Record (pos, closed, l) ->
        record sub env pos closed l
    | T.Nil pos -> nil env pos
    | T.Cons (pos, hd, tl) -> cons env pos (sub hd) (sub tl)
    | T.List (pos, l) ->
        list sub env pos l
    | T.Concat (pos, _, _)
    | T.Const (pos, Const.String _) ->
        begin
          try
            let s = to_string t in
            call env pos ["kernel"; "string"] "make" [
              Value.Term (Term.Const (Const.String s));
              Value.of_int (String.length s)
            ]
          with Not_a_string ->
            begin
              match t with
                | T.Concat (pos, v0, v1) ->
                    let v1 = sub v1 in
                    begin
                      try
                        let l = to_list v0 in
                        append sub env pos l v1
                      with Not_a_list ->
                        let v0 = sub v0 in
                        call env pos ["data"; "term"] "concat" [v0; v1]
                    end
                | _ ->
                    failwith "to_term"
            end
        end

let rec to_val env pos v =
  let (_, d) = Unionfind.find v in
  match d with
    | Unionfind.Fresh f -> Value.Var (fresh_to_var env.expr pos f)
    | Unionfind.Value (ExternVar v) -> Value.Var v
    | Unionfind.Value (Term t) -> to_term (to_val env pos) env t

let mk_call env (pos, t) = Eq.Call (pos, Compound.map (to_val env pos) t)

let mk_tok env (pos, to_consume, m, t) =
  Eq.Tok (pos, to_consume, to_val env pos m, Compound.map (to_val env pos) t)

let mk_eq env (pos, a, b) =
  Eq.Eq (pos, to_val env pos a, to_val env pos b)

let mk_ask env to_eq (ctx, pos, m, t, a) =
  let f (rev_new_args, ctx) v =
    let v' = Var.fresh () in
    (v' :: rev_new_args, String.Map.add v (of_extern_var v') ctx) in
  let (rev_new_args, ctx) = Array.fold_left f ([], ctx) t.Compound.args in
  let t' = Compound.mk (Compound.f t) (Array.of_list (List.rev rev_new_args)) in
  Eq.Ask (pos, to_val env pos m, t', to_eq (new_parent env.expr ctx pos) a)

let rec to_eq parent t =
  let env = {
    expr = create_expr_env parent;
    calls = [];
    asks = [];
    internal_agents = [];
  } in
  split env String.Map.empty t;
  let agent_calls = Eq.big_par (List.map (mk_call env) env.calls) in
  let agent_toks = Eq.big_par (List.map (mk_tok env) env.expr.toks) in
  let agent_asks = Eq.big_par (List.map (mk_ask env to_eq) env.asks) in
  let agent_eqs = Eq.big_par (List.map (mk_eq env) env.expr.eqs) in
  let internal_agents = Eq.big_par env.internal_agents in
  Eq.big_exists_pos env.expr.vars
    (Eq.big_par
       [agent_calls; agent_toks; internal_agents; agent_asks; agent_eqs])

let to_eq t = to_eq (fun v -> of_extern_var v) t
